@@ -17,15 +17,20 @@
 #include <errno.h>
 #include <grp.h>
 #include <json.h>
-#include <grp.h>
 #include <nss.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <time.h>
+#include <unistd.h>
+#include <sys/stat.h>
 
 #include <cstring>
+#include <cstdarg>
 #include <iostream>
 #include <sstream>
-#include "json_object.h"
+#include <fstream>
+
+#include <json_object.h>
 
 #if defined(__clang__) || __GNUC__ > 4 || \
     (__GNUC__ == 4 &&                     \
@@ -43,13 +48,78 @@
 using std::string;
 
 // Maximum number of retries for HTTP requests.
-const int kMaxRetries = 1;
+const int kMaxRetries = 3;
+
+// Backoff duration 1 sec between retries.
+const int kBackoffDuration = 1;
 
 // Regex for validating user names.
-const char kUserNameRegex[] = "^[a-zA-Z0-9._][a-zA-Z0-9._-]{0,31}$";
+static const char kUserNameRegex[] = "^[a-zA-Z0-9._][a-zA-Z0-9._-]{0,31}$";
+static const char kSudoersDir[] = "/var/google-sudoers.d/";
+static const char kUsersDir[] = "/var/google-users.d/";
 
 namespace oslogin_utils {
 
+// SysLog wraps syslog operations.
+class SysLog {
+ private:
+  // app is the application name or the application prefix.
+  const char *app;
+ public:
+  // Prints out an error to syslog.
+  void Error(const char *fmt, va_list args);
+
+  // Closes the file descriptor being used to write to sys logger.
+  void Close();
+
+  // Creates a SysLog instance specifying the ident. Additionally
+  // it carries an app identifier, so the syslog entries will look like:
+  // <<ident>>: <<app>>: <<Message>>
+  // For google_authorized_keys for example, it would look like:
+  // sshd: google_authorized_keys: <<Message>>
+  SysLog(const char *ident, const char *app);
+};
+
+static SysLog *logger = NULL;
+
+// ----------------- SysLog -------------------------
+SysLog::SysLog(const char *ident, const char *app) {
+  openlog(ident, LOG_PID|LOG_PERROR, LOG_DAEMON);
+  this->app = app;
+}
+
+void SysLog::Error(const char *fmt, va_list args) {
+  std::stringstream new_fmt;
+  new_fmt << this->app << ": " << fmt;
+  vsyslog(LOG_ERR, new_fmt.str().c_str(), args);
+}
+
+void SysLog::Close() {
+  closelog();
+}
+
+void SetupSysLog(const char *ident, const char *app) {
+  if (ident != NULL && logger == NULL) {
+    logger = new SysLog(ident, app);
+  }
+}
+
+void SysLogErr(const char *fmt, ...) {
+  if (logger != NULL) {
+    va_list args;
+    va_start(args, fmt);
+    logger->Error(fmt, args);
+    va_end(args);
+  }
+}
+
+void CloseSysLog() {
+  if (logger != NULL) {
+    logger->Close();
+    logger = NULL;
+  }
+}
+
 // ----------------- Buffer Manager -----------------
 
 BufferManager::BufferManager(char* buf, size_t buflen)
@@ -90,6 +160,7 @@ NssCache::NssCache(int cache_size)
     : cache_size_(cache_size),
       entry_cache_(cache_size),
       page_token_(""),
+      index_(0),
       on_last_page_(false) {}
 
 void NssCache::Reset() {
@@ -123,80 +194,108 @@ bool NssCache::GetNextGroup(BufferManager* buf, struct group* result, int* errno
 
 bool NssCache::LoadJsonUsersToCache(string response) {
   Reset();
+
   json_object* root = NULL;
   root = json_tokener_parse(response.c_str());
   if (root == NULL) {
     return false;
   }
+
+  bool ret = false;
+  int arraylen = 0;
+  json_object* login_profiles = NULL;
+
   // First grab the page token.
   json_object* page_token_object;
   if (json_object_object_get_ex(root, "nextPageToken", &page_token_object)) {
     page_token_ = json_object_get_string(page_token_object);
   } else {
-    return false;
+    goto cleanup;
   }
+
   // A page_token of 0 means we are done. This response will not contain any
   // login profiles.
   if (page_token_ == "0") {
     page_token_ = "";
     on_last_page_ = true;
-    return true;
+    ret = true;
+    goto cleanup;
   }
+
   // Now grab all of the loginProfiles.
-  json_object* login_profiles = NULL;
   if (!json_object_object_get_ex(root, "loginProfiles", &login_profiles)) {
-    return false;
+    goto cleanup;
   }
+
   if (json_object_get_type(login_profiles) != json_type_array) {
-    return false;
+    goto cleanup;
   }
-  int arraylen = json_object_array_length(login_profiles);
+
+  arraylen = json_object_array_length(login_profiles);
   if (arraylen == 0 || arraylen > cache_size_) {
-    return false;
+    goto cleanup;
   }
+
   for (int i = 0; i < arraylen; i++) {
     json_object* profile = json_object_array_get_idx(login_profiles, i);
     entry_cache_.push_back(json_object_to_json_string_ext(profile, JSON_C_TO_STRING_PLAIN));
   }
-  return true;
+  ret = true;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
-bool NssCache::LoadJsonGroupsToCache(string response) {
+bool NssCache::LoadJsonGroupsToCache(string response, int* errnop) {
   Reset();
+  *errnop = ENOENT;
+
   json_object* root = NULL;
   root = json_tokener_parse(response.c_str());
   if (root == NULL) {
     return false;
   }
+
+  bool ret = false;
+  int arraylen = 0;
+  json_object* groups = NULL;
+
   // First grab the page token.
   json_object* page_token_object;
   if (json_object_object_get_ex(root, "nextPageToken", &page_token_object)) {
     page_token_ = json_object_get_string(page_token_object);
   } else {
-    return false;
+    goto cleanup;
   }
   // A page_token of 0 for groups is different than for users. This is the last
-  // page, but it WILL contain groups.
+  // page, but it WILL contain groups if there are any.
   if (page_token_ == "0") {
     on_last_page_ = true;
     page_token_ = "";
   }
-  json_object* groups = NULL;
   if (!json_object_object_get_ex(root, "posixGroups", &groups)) {
-    return false;
+    // Valid JSON but no groups, set ENOMSG as a 'no groups' code.
+    *errnop = ENOMSG;
+    goto cleanup;
   }
   if (json_object_get_type(groups) != json_type_array) {
-    return false;
+    goto cleanup;
   }
-  int arraylen = json_object_array_length(groups);
+  arraylen = json_object_array_length(groups);
   if (arraylen == 0 || arraylen > cache_size_) {
-    return false;
+    goto cleanup;
   }
   for (int i = 0; i < arraylen; i++) {
     json_object* group = json_object_array_get_idx(groups, i);
     entry_cache_.push_back(json_object_to_json_string_ext(group, JSON_C_TO_STRING_PLAIN));
   }
-  return true;
+  ret = true;
+  *errnop = 0;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
 // Gets the next entry from the cache, refreshing as needed. Returns true if a
@@ -205,6 +304,8 @@ bool NssCache::LoadJsonGroupsToCache(string response) {
 //
 // * EINVAL  - current user entry was malformed in some way.
 // * ERANGE  - the page of results did not fit into the provided buffer.
+// * ENOMSG  - a 404 error was received when contacting the metadata server, indicating that
+//             OS Login is not enabled in the instance metadata.
 // * ENOENT  - a general failure to load the cache occurred. Behavior of retries
 //             following ENOENT is undefined.
 bool NssCache::NssGetpwentHelper(BufferManager* buf, struct passwd* result, int* errnop) {
@@ -217,8 +318,14 @@ bool NssCache::NssGetpwentHelper(BufferManager* buf, struct passwd* result, int*
     }
     string response;
     long http_code = 0;
-    if (!HttpGet(url.str(), &response, &http_code) || http_code != 200 ||
-        response.empty() || !LoadJsonUsersToCache(response)) {
+    bool status = HttpGet(url.str(), &response, &http_code);
+    // 404 means OS Login is not enabled.
+    if (http_code == 404) {
+      *errnop = ENOMSG;
+      return false;
+    }
+    // General failure to load the cache occurred.
+    if (!status || http_code != 200 || response.empty() || !LoadJsonUsersToCache(response)) {
       *errnop = ENOENT;
       return false;
     }
@@ -232,6 +339,8 @@ bool NssCache::NssGetpwentHelper(BufferManager* buf, struct passwd* result, int*
 //
 // * EINVAL  - current group entry was malformed in some way.
 // * ERANGE  - the page of results did not fit into the provided buffer.
+// * ENOMSG  - a 404 error was received when contacting the metadata server, indicating that
+//             OS Login is not enabled in the instance metadata.
 // * ENOENT  - a general failure to load the cache occurred. Behavior of retries
 //             following ENOENT is undefined.
 bool NssCache::NssGetgrentHelper(BufferManager* buf, struct group* result, int* errnop) {
@@ -244,11 +353,21 @@ bool NssCache::NssGetgrentHelper(BufferManager* buf, struct group* result, int*
     }
     string response;
     long http_code = 0;
-    if (!HttpGet(url.str(), &response, &http_code) || http_code != 200 ||
-        response.empty() || !LoadJsonGroupsToCache(response)) {
+    bool status = HttpGet(url.str(), &response, &http_code);
+    // 404 means OS Login is not enabled.
+    if (http_code == 404) {
+      *errnop = ENOMSG;
+      return false;
+    }
+    // Failed to make the request or empty response.
+    if (!status || http_code != 200 || response.empty()) {
       *errnop = ENOENT;
       return false;
     }
+    // General failure to load the cache occurred.
+    if (!LoadJsonGroupsToCache(response, errnop)) {
+      return false;
+    }
   }
 
   if (!HasNextEntry() || !GetNextGroup(buf, result, errnop)) {
@@ -276,6 +395,18 @@ size_t OnCurlWrite(void* buf, size_t size, size_t nmemb, void* userp) {
   return 0;
 }
 
+bool ShouldRetry(long http_code) {
+  if (http_code == 200) {
+    // Request returned successfully, no need to retry.
+    return false;
+  }
+  if (http_code == 404) {
+    // Metadata key does not exist, no point of retrying.
+    return false;
+  }
+  return true;
+}
+
 bool HttpDo(const string& url, const string& data, string* response, long* http_code) {
   if (response == NULL || http_code == NULL) {
     return false;
@@ -294,6 +425,10 @@ bool HttpDo(const string& url, const string& data, string* response, long* http_
       return false;
     }
     do {
+      // Apply backoff strategy before retrying.
+      if (retry_count > 0) {
+        sleep(kBackoffDuration);
+      }
       response_stream.str("");
       response_stream.clear();
       curl_easy_setopt(curl, CURLOPT_HTTPHEADER, header_list);
@@ -312,7 +447,7 @@ bool HttpDo(const string& url, const string& data, string* response, long* http_
         return false;
       }
       curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, http_code);
-    } while (retry_count++ < kMaxRetries && *http_code == 500);
+    } while (retry_count++ < kMaxRetries && ShouldRetry(*http_code));
     curl_slist_free_all(header_list);
   }
   *response = response_stream.str();
@@ -383,6 +518,7 @@ bool ValidatePasswd(struct passwd* result, BufferManager* buf, int* errnop) {
   if (!buf->AppendString("", &result->pw_gecos, errnop)) {
     return false;
   }
+
   return true;
 }
 
@@ -395,19 +531,26 @@ bool ParseJsonToUsers(const string& json, std::vector<string>* result) {
     return false;
   }
 
+  bool ret = false;
+
   json_object* users = NULL;
   if (!json_object_object_get_ex(root, "usernames", &users)) {
-    return false;
+    ret = true; // means no users, not invalid.
+    goto cleanup;
   }
   if (json_object_get_type(users) != json_type_array) {
-    return false;
+    goto cleanup;
   }
   for (int idx=0; idx < (int)json_object_array_length(users); idx++) {
     json_object* user = json_object_array_get_idx(users, idx);
     const char* username = json_object_get_string(user);
     result->push_back(string(username));
   }
-  return true;
+  ret = true;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
 bool ParseJsonToGroups(const string& json, std::vector<Group>* result) {
@@ -417,94 +560,125 @@ bool ParseJsonToGroups(const string& json, std::vector<Group>* result) {
     return false;
   }
 
+  bool ret = false;
+
   json_object* groups = NULL;
   if (!json_object_object_get_ex(root, "posixGroups", &groups)) {
-    return false;
+    goto cleanup;
   }
   if (json_object_get_type(groups) != json_type_array) {
-    return false;
+    goto cleanup;
   }
   for (int idx = 0; idx < (int)json_object_array_length(groups); idx++) {
     json_object* group = json_object_array_get_idx(groups, idx);
 
     json_object* gid;
     if (!json_object_object_get_ex(group, "gid", &gid)) {
-      return false;
+      goto cleanup;
     }
 
     json_object* name;
     if (!json_object_object_get_ex(group, "name", &name)) {
-      return false;
+      goto cleanup;
     }
 
     Group g;
     g.gid = json_object_get_int64(gid);
+
     // get_int64 will confusingly return 0 if the string can't be converted to
     // an integer. We can't rely on type check as it may be a string in the API.
+    // Also 0 is invalid because it creates a 'root group'.
     if (g.gid == 0) {
-      return false;
+      goto cleanup;
     }
+
     g.name = json_object_get_string(name);
     if (g.name == "") {
-      return false;
+      goto cleanup;
     }
 
     result->push_back(g);
   }
-  return true;
+  ret = true;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
-bool ParseJsonToGroup(const string& json, struct group* result, BufferManager* buf, int* errnop) {
+bool ParseJsonToGroup(const string& json, struct group* result, BufferManager*
+                      buf, int* errnop) {
+  *errnop = EINVAL;
+  int gr_gid = 65535;
+
   json_object* group = NULL;
   group = json_tokener_parse(json.c_str());
-  if (group== NULL) {
-    *errnop = EINVAL;
+  if (group == NULL) {
     return false;
   }
 
+  bool ret = false;
+
   json_object* gid;
   if (!json_object_object_get_ex(group, "gid", &gid)) {
-    *errnop = EINVAL;
-    return false;
+    goto cleanup;
   }
 
   json_object* name;
   if (!json_object_object_get_ex(group, "name", &name)) {
-    *errnop = EINVAL;
-    return false;
+    goto cleanup;
   }
 
-  result->gr_gid = json_object_get_int64(gid);
-  // TODO ValidateGroup
-  buf->AppendString("", &result->gr_passwd, errnop);
-  return buf->AppendString((char*)json_object_get_string(name), &result->gr_name, errnop);
+  if ((gr_gid = json_object_get_int64(gid)) == 0) {
+    goto cleanup;
+  }
+
+  result->gr_gid = gr_gid;
+  if (!buf->AppendString("", &result->gr_passwd, errnop))
+    goto cleanup;
+  if (!buf->AppendString(json_object_get_string(name), &result->gr_name,
+                         errnop))
+    goto cleanup;
+
+  *errnop = 0;
+  ret = true;
+
+cleanup:
+  json_object_put(group);
+  return ret;
 }
 
 std::vector<string> ParseJsonToSshKeys(const string& json) {
   std::vector<string> result;
+  json_object* ssh_public_keys = NULL;
+
   json_object* root = NULL;
   root = json_tokener_parse(json.c_str());
   if (root == NULL) {
     return result;
   }
+
   // Locate the sshPublicKeys object.
   json_object* login_profiles = NULL;
   if (!json_object_object_get_ex(root, "loginProfiles", &login_profiles)) {
-    return result;
+    goto cleanup;
   }
   if (json_object_get_type(login_profiles) != json_type_array) {
-    return result;
+    goto cleanup;
   }
   login_profiles = json_object_array_get_idx(login_profiles, 0);
 
-  json_object* ssh_public_keys = NULL;
   if (!json_object_object_get_ex(login_profiles, "sshPublicKeys", &ssh_public_keys)) {
-    return result;
+    goto cleanup;
   }
 
   if (json_object_get_type(ssh_public_keys) != json_type_object) {
-    return result;
+    goto cleanup;
   }
+
+  {
+  // Extra braces to indicate scope of key, obj below to compiler. Otherwise
+  // g++ complains that `goto` bypasses initializers.
   json_object_object_foreach(ssh_public_keys, key, obj) {
     (void)(key);
     if (json_object_get_type(obj) != json_type_object) {
@@ -519,7 +693,7 @@ std::vector<string> ParseJsonToSshKeys(const string& json) {
         if (val_type != json_type_string) {
           continue;
         }
-        key_to_add = (char*)json_object_get_string(val);
+        key_to_add = json_object_get_string(val);
       }
       if (string_key == "expirationTimeUsec") {
         if (val_type == json_type_int || val_type == json_type_string) {
@@ -537,33 +711,98 @@ std::vector<string> ParseJsonToSshKeys(const string& json) {
       result.push_back(key_to_add);
     }
   }
+  }
+
+cleanup:
+  json_object_put(root);
+  return result;
+}
+
+std::vector<string> ParseJsonToSshKeysSk(const string& json) {
+  std::vector<string> result;
+  json_object* security_keys = NULL;
+
+  json_object* root = NULL;
+  root = json_tokener_parse(json.c_str());
+  if (root == NULL) {
+    return result;
+  }
+
+  // Locate the securityKeys array.
+  json_object* login_profiles = NULL;
+  if (!json_object_object_get_ex(root, "loginProfiles", &login_profiles)) {
+    goto cleanup;
+  }
+  if (json_object_get_type(login_profiles) != json_type_array) {
+    goto cleanup;
+  }
+
+  login_profiles = json_object_array_get_idx(login_profiles, 0);
+
+  if (!json_object_object_get_ex(login_profiles, "securityKeys", &security_keys)) {
+    goto cleanup;
+  }
+  if (json_object_get_type(security_keys) != json_type_array) {
+    goto cleanup;
+  }
+
+  {
+    size_t number_of_keys = 0;
+    size_t idx;
+    json_object* security_key = NULL;
+    json_object* public_key = NULL;
+    string key_to_add = "";
+
+    number_of_keys = json_object_array_length(security_keys);
+    for (idx = 0; idx < number_of_keys; idx++) {
+      security_key = json_object_array_get_idx(security_keys, idx);
+      if (json_object_get_type(security_key) != json_type_object) {
+        goto cleanup;
+      }
+      if (!json_object_object_get_ex(security_key, "publicKey", &public_key)) {
+        goto cleanup;
+      }
+
+      key_to_add = json_object_get_string(public_key);
+      result.push_back(key_to_add);
+      key_to_add.clear();
+    }
+  }
+
+cleanup:
+  json_object_put(root);
   return result;
 }
 
 bool ParseJsonToPasswd(const string& json, struct passwd* result, BufferManager*
                        buf, int* errnop) {
+  *errnop = EINVAL;
   json_object* root = NULL;
-  root = json_tokener_parse(json.c_str());
+  json_object* origroot = NULL;
+
+  origroot = root = json_tokener_parse(json.c_str());
   if (root == NULL) {
-    *errnop = EINVAL;
     return false;
   }
+
+  bool ret = false;
+  json_object* posix_accounts = NULL;
+
   json_object* login_profiles = NULL;
   // If this is called from getpwent_r, loginProfiles won't be in the response.
   if (json_object_object_get_ex(root, "loginProfiles", &login_profiles)) {
     if (json_object_get_type(login_profiles) != json_type_array) {
-      return false;
+      goto cleanup;
     }
+    // This overwrites root but we still have origroot for cleanup;
     root = json_object_array_get_idx(login_profiles, 0);
   }
   // Locate the posixAccounts object.
-  json_object* posix_accounts = NULL;
   if (!json_object_object_get_ex(root, "posixAccounts", &posix_accounts)) {
-    *errnop = EINVAL;
-    return false;
+    goto cleanup;
   }
   if (json_object_get_type(posix_accounts) != json_type_array) {
-    return false;
+    goto cleanup;
   }
   posix_accounts = json_object_array_get_idx(posix_accounts, 0);
 
@@ -577,9 +816,11 @@ bool ParseJsonToPasswd(const string& json, struct passwd* result, BufferManager*
 
   // Iterate through the json response and populate the passwd struct.
   if (json_object_get_type(posix_accounts) != json_type_object) {
-    *errnop = EINVAL;
-    return false;
+    goto cleanup;
   }
+  {
+  // Extra braces to indicate scope of key, obj below to compiler. Otherwise
+  // g++ complains that `goto` bypasses initializers.
   json_object_object_foreach(posix_accounts, key, val) {
     int val_type = json_object_get_type(val);
     // Convert char* to c++ string for easier comparison.
@@ -589,12 +830,10 @@ bool ParseJsonToPasswd(const string& json, struct passwd* result, BufferManager*
       if (val_type == json_type_int || val_type == json_type_string) {
         result->pw_uid = (uint32_t)json_object_get_int64(val);
         if (result->pw_uid == 0) {
-          *errnop = EINVAL;
-          return false;
+          goto cleanup;
         }
       } else {
-        *errnop = EINVAL;
-        return false;
+        goto cleanup;
       }
     } else if (string_key == "gid") {
       if (val_type == json_type_int || val_type == json_type_string) {
@@ -604,40 +843,41 @@ bool ParseJsonToPasswd(const string& json, struct passwd* result, BufferManager*
           result->pw_gid = result->pw_uid;
         }
       } else {
-        *errnop = EINVAL;
-        return false;
+        goto cleanup;
       }
     } else if (string_key == "username") {
       if (val_type != json_type_string) {
-        *errnop = EINVAL;
-        return false;
+        goto cleanup;
       }
-      if (!buf->AppendString((char*)json_object_get_string(val),
+      if (!buf->AppendString(json_object_get_string(val),
                              &result->pw_name, errnop)) {
-        return false;
+        goto cleanup;
       }
     } else if (string_key == "homeDirectory") {
       if (val_type != json_type_string) {
-        *errnop = EINVAL;
-        return false;
+        goto cleanup;
       }
-      if (!buf->AppendString((char*)json_object_get_string(val),
+      if (!buf->AppendString(json_object_get_string(val),
                              &result->pw_dir, errnop)) {
-        return false;
+        goto cleanup;
       }
     } else if (string_key == "shell") {
       if (val_type != json_type_string) {
-        *errnop = EINVAL;
-        return false;
+        goto cleanup;
       }
-      if (!buf->AppendString((char*)json_object_get_string(val),
+      if (!buf->AppendString(json_object_get_string(val),
                              &result->pw_shell, errnop)) {
-        return false;
+        goto cleanup;
       }
     }
   }
+  }
+  *errnop = 0;
+  ret = ValidatePasswd(result, buf, errnop);
 
-  return ValidatePasswd(result, buf, errnop);
+cleanup:
+  json_object_put(origroot);
+  return ret;
 }
 
 bool AddUsersToGroup(std::vector<string> users, struct group* result,
@@ -672,22 +912,28 @@ bool ParseJsonToEmail(const string& json, string* email) {
   if (root == NULL) {
     return false;
   }
+
+  bool ret = false;
+  json_object* json_email = NULL;
+
   // Locate the email object.
   json_object* login_profiles = NULL;
   if (!json_object_object_get_ex(root, "loginProfiles", &login_profiles)) {
-    return false;
+    goto cleanup;
   }
   if (json_object_get_type(login_profiles) != json_type_array) {
-    return false;
+    goto cleanup;
   }
   login_profiles = json_object_array_get_idx(login_profiles, 0);
-  json_object* json_email = NULL;
   if (!json_object_object_get_ex(login_profiles, "name", &json_email)) {
-    return false;
+    goto cleanup;
   }
-
+  ret = true;
   *email = json_object_get_string(json_email);
-  return true;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
 bool ParseJsonToSuccess(const string& json) {
@@ -698,59 +944,70 @@ bool ParseJsonToSuccess(const string& json) {
   }
   json_object* success = NULL;
   if (!json_object_object_get_ex(root, "success", &success)) {
+    json_object_put(root);
     return false;
   }
-  return (bool)json_object_get_boolean(success);
+  bool ret = (bool)json_object_get_boolean(success);
+  json_object_put(root);
+  return ret;
 }
 
 bool ParseJsonToKey(const string& json, const string& key, string* response) {
   json_object* root = NULL;
-  json_object* json_response = NULL;
-  const char* c_response;
-
   root = json_tokener_parse(json.c_str());
   if (root == NULL) {
     return false;
   }
 
+  bool ret = false;
+  json_object* json_response = NULL;
+  const char* c_response = NULL;
+
+
   if (!json_object_object_get_ex(root, key.c_str(), &json_response)) {
-    return false;
+    goto cleanup;
   }
 
   if (!(c_response = json_object_get_string(json_response))) {
-    return false;
+    goto cleanup;
   }
 
   *response = c_response;
-  return true;
+  ret = true;
+
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
 bool ParseJsonToChallenges(const string& json, std::vector<Challenge>* challenges) {
   json_object* root = NULL;
-
   root = json_tokener_parse(json.c_str());
   if (root == NULL) {
     return false;
   }
 
+  bool ret = false;
+  json_object* challengeId = NULL;
+  json_object* challengeType = NULL;
+  json_object* challengeStatus = NULL;
   json_object* jsonChallenges = NULL;
   if (!json_object_object_get_ex(root, "challenges", &jsonChallenges)) {
-    return false;
+    goto cleanup;
   }
 
-  json_object *challengeId, *challengeType, *challengeStatus = NULL;
   for (int i = 0; i < (int)json_object_array_length(jsonChallenges); ++i) {
     if (!json_object_object_get_ex(json_object_array_get_idx(jsonChallenges, i),
                                    "challengeId", &challengeId)) {
-      return false;
+      goto cleanup;
     }
     if (!json_object_object_get_ex(json_object_array_get_idx(jsonChallenges, i),
                                    "challengeType", &challengeType)) {
-      return false;
+      goto cleanup;
     }
     if (!json_object_object_get_ex(json_object_array_get_idx(jsonChallenges, i),
                                    "status", &challengeStatus)) {
-      return false;
+      goto cleanup;
     }
     Challenge challenge;
     challenge.id = json_object_get_int(challengeId);
@@ -759,74 +1016,15 @@ bool ParseJsonToChallenges(const string& json, std::vector<Challenge>* challenge
 
     challenges->push_back(challenge);
   }
+  ret = true;
 
-  return true;
+cleanup:
+  json_object_put(root);
+  return ret;
 }
 
 // ----------------- OS Login functions -----------------
 
-// TODO: this function reads all groups comparing names or gids; it should be
-// replaced by groups?groupname= lookup when this is available.
-bool FindGroup(struct group* result, BufferManager* buf, int* errnop) {
-  if (result->gr_name == NULL && result->gr_gid == 0) {
-    // Nobody told me what to find.
-    return false;
-  }
-  std::stringstream url;
-  std::vector<Group> groups;
-
-  string response;
-  long http_code;
-  string pageToken = "";
-
-  do {
-    url.str("");
-    url << kMetadataServerUrl << "groups";
-    if (pageToken != "")
-      url << "?pageToken=" << pageToken;
-
-    response.clear();
-    http_code = 0;
-    if (!HttpGet(url.str(), &response, &http_code) || http_code != 200 ||
-        response.empty()) {
-      *errnop = EAGAIN;
-      return false;
-    }
-
-    if (!ParseJsonToKey(response, "nextPageToken", &pageToken)) {
-      pageToken = "";
-    }
-
-    groups.clear();
-    if (!ParseJsonToGroups(response, &groups) || groups.empty()) {
-      *errnop = ENOENT;
-      return false;
-    }
-
-    // Check for a match.
-    for (int i = 0; i < (int) groups.size(); i++) {
-      Group el = groups[i];
-      if ((result->gr_name != NULL) && (string(result->gr_name) == el.name)) {
-        // Set the name even though it matches because the final string must
-        // be stored in the provided buffer.
-        if (!buf->AppendString(el.name, &result->gr_name, errnop)) {
-          return false;
-        }
-        result->gr_gid = el.gid;
-        return true;
-      }
-      if ((result->gr_gid != 0) && (result->gr_gid == el.gid)) {
-        if (!buf->AppendString(el.name, &result->gr_name, errnop)) {
-          return false;
-        }
-        return true;
-      }
-    }
-  } while (pageToken != "0");
-  // Not found.
-  *errnop = ENOENT;
-  return false;
-}
 
 bool GetGroupsForUser(string username, std::vector<Group>* groups, int* errnop) {
   string response;
@@ -846,13 +1044,13 @@ bool GetGroupsForUser(string username, std::vector<Group>* groups, int* errnop)
   std::stringstream url;
 
   long http_code;
-  string pageToken = "";
+  string pageToken ("");
 
   do {
     url.str("");
     url << kMetadataServerUrl << "groups?email=" << email;
     if (pageToken != "")
-      url << "?pageToken=" << pageToken;
+      url << "&pagetoken=" << pageToken;
 
     response.clear();
     http_code = 0;
@@ -862,29 +1060,96 @@ bool GetGroupsForUser(string username, std::vector<Group>* groups, int* errnop)
       return false;
     }
 
-    if (!ParseJsonToKey(response, "pageToken", &pageToken)) {
-      pageToken = "";
+    if (!ParseJsonToKey(response, "nextPageToken", &pageToken)) {
+      *errnop = ENOENT;
+      return false;
     }
 
     if (!ParseJsonToGroups(response, groups)) {
       *errnop = ENOENT;
       return false;
     }
-  } while (pageToken != "");
+  } while (pageToken != "0");
+  return true;
+}
+
+bool GetGroupByName(string name, struct group* result, BufferManager* buf, int* errnop) {
+  std::stringstream url;
+  std::vector<Group> groups;
+
+  string response;
+  long http_code;
+
+  url.str("");
+  url << kMetadataServerUrl << "groups?groupname=" << name;
+
+  response.clear();
+  http_code = 0;
+  if (!HttpGet(url.str(), &response, &http_code) || http_code != 200 ||
+      response.empty()) {
+    *errnop = EAGAIN;
+    return false;
+  }
+
+  groups.clear();
+  if (!ParseJsonToGroups(response, &groups) || groups.empty() || groups.size() != 1) {
+    *errnop = ENOENT;
+    return false;
+  }
+
+  Group el = groups[0];
+  result->gr_gid = el.gid;
+  if (!buf->AppendString(el.name, &result->gr_name, errnop)) {
+    return false;
+  }
+
+  return true;
+}
+
+bool GetGroupByGID(int gid, struct group* result, BufferManager* buf, int* errnop) {
+  std::stringstream url;
+  std::vector<Group> groups;
+
+  string response;
+  long http_code;
+
+  url.str("");
+  url << kMetadataServerUrl << "groups?gid=" << gid;
+
+  response.clear();
+  http_code = 0;
+  if (!HttpGet(url.str(), &response, &http_code) || http_code != 200 ||
+      response.empty()) {
+    *errnop = EAGAIN;
+    return false;
+  }
+
+  groups.clear();
+  if (!ParseJsonToGroups(response, &groups) || groups.empty() || groups.size() != 1) {
+    *errnop = ENOENT;
+    return false;
+  }
+
+  Group el = groups[0];
+  result->gr_gid = el.gid;
+  if (!buf->AppendString(el.name, &result->gr_name, errnop)) {
+    return false;
+  }
+
   return true;
 }
 
 bool GetUsersForGroup(string groupname, std::vector<string>* users, int* errnop) {
   string response;
   long http_code;
-  string pageToken = "";
+  string pageToken ("");
   std::stringstream url;
 
   do {
     url.str("");
     url << kMetadataServerUrl << "users?groupname=" << groupname;
     if (pageToken != "")
-      url << "?pageToken=" << pageToken;
+      url << "&pagetoken=" << pageToken;
 
     response.clear();
     http_code = 0;
@@ -894,10 +1159,10 @@ bool GetUsersForGroup(string groupname, std::vector<string>* users, int* errnop)
       return false;
     }
     if (!ParseJsonToKey(response, "nextPageToken", &pageToken)) {
-      pageToken = "";
+      *errnop = EINVAL;
+      return false;
     }
     if (!ParseJsonToUsers(response, users)) {
-    // TODO: what if there are no users? add a test.
       *errnop = EINVAL;
       return false;
     }
@@ -905,21 +1170,31 @@ bool GetUsersForGroup(string groupname, std::vector<string>* users, int* errnop)
   return true;
 }
 
-bool GetUser(const string& username, string* response) {
+bool MDSGetUser(const string& username, bool security_key, string* response) {
   std::stringstream url;
   url << kMetadataServerUrl << "users?username=" << UrlEncode(username);
 
+  if (security_key) {
+    url << "&view=securityKey";
+  }
+
   long http_code = 0;
   if (!HttpGet(url.str(), response, &http_code) || response->empty() ||
       http_code != 200) {
     return false;
   }
+
   return true;
 }
 
+bool GetUser(const string& username, string* response) {
+  return MDSGetUser(username, false, response);
+}
+
 bool StartSession(const string& email, string* response) {
   bool ret = true;
-  struct json_object *jobj, *jarr;
+  json_object* jobj = NULL;
+  json_object* jarr = NULL;
 
   jarr = json_object_new_array();
   json_object_array_add(jarr, json_object_new_string(INTERNAL_TWO_FACTOR));
@@ -951,7 +1226,8 @@ bool StartSession(const string& email, string* response) {
 
 bool ContinueSession(bool alt, const string& email, const string& user_token, const string& session_id, const Challenge& challenge, string* response) {
   bool ret = true;
-  struct json_object *jobj, *jresp;
+  json_object* jobj = NULL;
+  json_object* jresp = NULL;
 
   jobj = json_object_new_object();
   json_object_object_add(jobj, "email", json_object_new_string(email.c_str()));
@@ -989,4 +1265,146 @@ bool ContinueSession(bool alt, const string& email, const string& user_token, co
 
   return ret;
 }
+
+static bool ApplyPolicy(const char *user_name, string email, const char *policy, struct AuthOptions opts) {
+  std::stringstream url;
+  url << kMetadataServerUrl << "authorize?email=" << UrlEncode(email) << "&policy=" << policy;
+
+  // Don't try to add fingerprint parameter to policy call if we don't have it.
+  if (opts.fp_len > 0) {
+    url << "&fingerprint=" << opts.fingerprint;
+  }
+
+  string response;
+  long http_code = 0;
+  // Invalid user, just leave from here - the principal will not be allowed/authorized.
+  if (!HttpGet(url.str(), &response, &http_code)) {
+    SysLogErr("Failed to validate organization user %s has login permission.", user_name);
+    return false;
+  }
+
+  if (http_code != 200) {
+    SysLogErr("Failed to validate organization user %s has login permission, "
+              "got HTTP response code: %lu", user_name, http_code);
+    return false;
+  }
+
+  if (!ParseJsonToSuccess(response)) {
+    SysLogErr("Organization user %s does not have login permission.", user_name);
+    return false;
+  }
+
+  return true;
+}
+
+static bool FileExists(const char *file_path) {
+  struct stat buff;
+  return !stat(file_path, &buff);
+}
+
+static bool CreateGoogleUserFile(string users_filename) {
+  std::ofstream users_file;
+
+  users_file.open(users_filename.c_str());
+
+  if (!users_file.is_open()) {
+    // If we can't open the file (meaning we can't create it) we should report failure.
+    return false;
+  }
+
+  // We are only creating the file so we could just close it here.
+  users_file.close();
+
+  chown(users_filename.c_str(), 0, 0);
+  chmod(users_filename.c_str(), S_IRUSR | S_IWUSR | S_IRGRP);
+  return true;
+}
+
+static bool CreateGoogleSudoersFile(string sudoers_filename, const char *user_name) {
+  std::ofstream sudoers_file;
+
+  sudoers_file.open(sudoers_filename.c_str());
+
+  if (!sudoers_file.is_open()) {
+    // If we can't open the file (meaning we can't create it) we should report failure.
+    return false;
+  }
+
+  sudoers_file << user_name << " ALL=(ALL) NOPASSWD: ALL\n";
+  sudoers_file.close();
+
+  chown(sudoers_filename.c_str(), 0, 0);
+  chmod(sudoers_filename.c_str(), S_IRUSR | S_IRGRP);
+  return true;
+}
+
+bool AuthorizeUser(const char *user_name, struct AuthOptions opts, string *user_response) {
+  bool users_file_exists, sudoers_exists;
+  string email, users_filename, sudoers_filename;
+
+  users_file_exists = sudoers_exists = false;
+
+  if (!ValidateUserName(user_name)) {
+    return false;
+  }
+
+  // Call MDS "users?username=" endpoint.
+  if (!MDSGetUser(user_name, opts.security_key, user_response)) {
+    return false;
+  }
+
+  if (!ParseJsonToEmail(*user_response, &email) || email.empty()) {
+    return false;
+  }
+
+  users_filename = kUsersDir;
+  users_filename.append(user_name);
+  users_file_exists = FileExists(users_filename.c_str());
+
+  if (!ApplyPolicy(user_name, email, "login", opts)) {
+    // Couldn't apply "login" policy for user in question, log it and deny.
+    SysLogErr("Could not grant access to organization user: %s.", user_name);
+    if (users_file_exists) {
+      remove(users_filename.c_str());
+    }
+    return false;
+  }
+
+  if (!users_file_exists && !CreateGoogleUserFile(users_filename)) {
+    // If we can't create users file we can't grant access, log it and deny.
+    SysLogErr("Failed to create user's file.");
+    return false;
+  }
+
+  sudoers_filename = kSudoersDir;
+  sudoers_filename.append(user_name);
+  sudoers_exists = FileExists(sudoers_filename.c_str());
+
+  if (ApplyPolicy(user_name, email, "adminLogin", opts)) {
+    // Best effort creating sudoers file, if we fail log it and grant access.
+    if (!sudoers_exists && !CreateGoogleSudoersFile(sudoers_filename, user_name)) {
+      SysLogErr("Could not grant sudo permissions to organization user %s."
+                " Sudoers file %s is not writable.", user_name, sudoers_filename.c_str());
+    }
+  } else {
+    remove(sudoers_filename.c_str());
+  }
+
+  return true;
+}
+
+const char *FileName(const char *file_path) {
+  int res_start = 0;
+  for (int i = 0; file_path[i] != '\0'; i++) {
+    if (file_path[i] == '/') {
+      res_start = i;
+    }
+  }
+
+  if (res_start > 0) {
+    return file_path + res_start + 1;
+  }
+
+  return file_path;
+}
 }  // namespace oslogin_utils
